#!/usr/local/bin/perl

use Getopt::Long;
use PropertyList;
use TempFileNames;
use Set;
use Postscript;

sub optionOn { my ($option) = @_;
	return 0 if (!defined($option));
	return 1 if ($option == 1);	# from a command line option
	return uc($option) eq 'YES';
}

sub dhashdepth { my ($h) = @_;
	my $depth = 0;
	for( ; ref($h) eq 'HASH'; $h = $h->{(keys %{$h})[0]}) { $depth++; }
	return $depth;
}
sub dhash { my ($h, @keys) = @_;
	foreach $k (@keys) { $h = $h->{$k}; }
	return $h;
}
sub setdhash { my ($h, $v, @keys) = @_;
	for (my $i = 0; $i < $#keys; $i++) {
		$h = defined($h->{$keys[$i]})? $h->{$keys[$i]}: ($h->{$keys[$i]} = {});
	}
	$h->{$keys[$#keys]} = $v;
}

sub scrutinizeInteraction { my ($c, $model, $field) = @_;
	my @lines = split(/\n/o, $model);
	my ($mx, $mn) = ({}, {});

	shift(@lines);
	foreach $line (@lines) {
		my @fields = split(/\t/o, $line);
		my @loci = sort ($fields[0] =~ m{(\S+?):\d+,?}og);
		my @alleles = sort ($fields[0] =~ m{(\S+?:\d+),?}og);

		if (defined($c->{single})) {
			setdhash($mx, $fields[$field], @alleles) if ($fields[$field] > 1);
			setdhash($mn, $fields[$field], @alleles) if ($fields[$field] < 1);
		} else {
			setdhash($mx, $fields[$field], @loci)
				if (!defined(dhash($mx, @loci)) || $fields[$field] > dhash($mx, @loci));
			setdhash($mn, $fields[$field], @loci)
				if (!defined(dhash($mn, @loci)) || $fields[$field] < dhash($mn, @loci));
		}
	}
	return ($mx, $mn);
}

sub printMatrices { my ($mx, $mn) = @_;
	# <b> perl bug prevents the expression: sort toList(...)
	my @loci = toList(union(keysOfDictLevel($mx, 1), keysOfDictLevel($mn, 1)));
	@loci = sort @loci;

	print join("\t", '', @loci), "\n";
	foreach $l1 (@loci) {
		print $l1;
		foreach $l2 (@loci) {
			print "\t", ($l2 eq $l1? '-1'
				: (($l2 cmp $l1) > 0 ? $mx->{$l1}{$l2}: $mn->{$l2}{$l1}));
		}
		print "\n";
	}

	print join(":", @{$loci}), "\n";
#toList(union([keys %{$mx}], [keys %{$mn}]));

};


sub fractionFromOR { my ($c, $or) = @_;
	my ($this, $min, $max) = optionOn($c->{useLogScale})
	  ? ((!$or? 0: log($or)), log($c->{minOR}), log($c->{maxOR}))
	  :	($or, $c->{minOR}, $c->{maxOR});
	return $this < $min? 0
		: ($this > $max? 1: (($this - $min) / ($max - $min)) );
}

sub printDotXY { my ($ps, $c, $x, $y, $f) = @_;
	my $cw2 = $c->{cellWidth}/2;
	$ps->arcfill($x + $cw2, $y + $cw2, $c->{dotMax} * abs($f)) if ($f);
}

sub printColorDotXY { my ($ps, $c, $x, $y, $f) = @_;
	return if (!$f);
	my $v = valueFromFraction($c, $f, $f < 0);
	my $cw2 = $c->{cellWidth}/2;

	$ps->setInterpolatedColor($f < 0? $c->{colorsORsmallerOne}: $c->{colorsORgreaterOne}, $v);
	$ps->arcfill($x + $cw2, $y + $cw2, $c->{dotMax} * abs($f)) if ($f);
}
sub printColorDotXY_old { my ($ps, $c, $x, $y, $f) = @_;
	return if (!$f);
	my $color = $f < 0? $c->{dots}{colorl}: $c->{dots}{colorh};
	$ps->setBlendedColorFromColorsAndFraction({ r => 1, g => 1, b => 1}, $color, abs($f));
	$ps->rectfill($x, $y, $c->{cellWidth}, $c->{cellWidth});
}

# Two ranges are indicated by the sign of $f (may be ingored -> abs($f))
sub printDot { my ($ps, $x, $y, $f, $cfg, $extraY) = @_;
	$cfg->{dots}{fct}($ps, $cfg,
		$cfg->{leftBorder} + $x * $cfg->{cellWidth},
		$cfg->{bottomBorder} + $extraY + $y * $cfg->{cellWidth},
		$f
	);
}

sub valueFromFraction { my ($c, $f, $i) = @_;	#config, fraction, inverse-flag
	return !$i? $c->{minOR}: (1/$c->{minOR}) if (!$f);
	return !$i? $c->{maxOR}: (1/$c->{maxOR}) if ($f == 1);
	my $v;
	if (optionOn($c->{useLogScale})) {
		$v = exp($f * log($c->{maxOR}) + ($f - 1) * log($c->{minOR}) );
	} else { $v = ($c->{maxOR} - $c->{minOR}) * $f + $c->{minOR}; }
	return !$i? $v: (1/$v);
}

sub drawLegendYI { my ($ps, $c, $orc, $lociCount, $yBase, $inverse) = @_;
	my $l = $c->{legend};
	my ($cw, $inset, $is) = ($c->{cellWidth}, $l->{inset}, $l->{interspace});
	my ($x, $y, $i) = ($c->{leftBorder} + $cw * ($lociCount + 1) + $l->{x},
		$c->{bottomBorder} + $yBase);

	$ps->newpath();
	$ps->gsave();
	for ($i = 0; $i < $l->{count}; $i++) {
		$c->{dots}{fct}($ps, $c, $x, $y + $i * $is + $is/2 - $cw/2,
		($i / ($l->{count} - 1)) * ($inverse? -1: 1));
	}
	$ps->grestore();

	$ps->setfont($l->{font}{name}, $l->{font}{size});
	for ($i = 0; $i < $l->{count}; $i++) {
		my $prefix = $inverse? ((!$i)? '> ': ($i == $l->{count} - 1)? '< ': '')
			: ((!$i)? '< ': ($i == $l->{count} - 1)? '> ': '');
		my $v = valueFromFraction($orc, $i / ($l->{count} - 1), $inverse);
		$ps->ycenteredShow($prefix.sprintf("%.2f", $v),
			$x + $inset, $y + $i * $is + $is/2);
	}
	if (uc($c->{rectAroundLegend}) eq 'YES') {
		$ps->rectstroke($x - $inset, $y - $inset,
			$l->{width} + 2 * $inset,
			$l->{count} * $is + 2 * $inset
		)
	}
}

sub drawLegend { my ($ps, $c, $lociCount) = @_;
	my $l = $c->{legend};
	my ($inset, $is) = ($l->{inset}, $l->{interspace});
	drawLegendYI($ps, $c, $c->{dots}, $lociCount, 0, 0);
	drawLegendYI($ps, $c, $c->{dots}, $lociCount, $l->{count} * $is + 4 * $inset, 1);
}

$TextWidth = 100;
sub printPostscript2 { my ($ps, $c, $mx, $mn) = @_;
	my @loci = toList(union(keysOfDictLevel($mx, 1), keysOfDictLevel($mn, 1)));
	@loci = @allLoci if ($allLoci);
	@loci = sort { sortTextNumber($a, $b) } @loci;
	print STDERR "Count of Loci ", int(@loci), "\n";
	my $cw = $c->{cellWidth};

	$ps->setgray(0);
	if (defined($inverseColors)) {
		$ps->showRyC($loci[int(@loci) - $x - 1], $c->{leftBorder},
			$c->{bottomBorder} + $x * $cw + $cw/2);
		$ps->rectfill(0, 0,
			$c->{leftBorder} + int(@loci) * $cw + $TextWidth,
			$c->{bottomBorder} + int(@loci) * $cw + $TextWidth);
		$ps->setgray(1);
	}

	my ($x, $y);
	my ($cellWidth, $bottom, $left) = ($c->{cellWidth}, $c->{bottomBorder},
		$c->{leftBorder});

	# print grid
	if (!optionOn($c->{noGrid})) {
		$ps->gsave();
		$ps->setgray(.9);
		for ($x = 0; $x < @loci; $x++) {
			for ($y = 0; $y < @loci; $y++) {
				if (!(($x) % 2)) {
					$ps->rectfill($left + $x * $cellWidth, $bottom,
						$cellWidth, int(@loci) * $cellWidth);
				}
				if (!(($y) % 4)) {
					$ps->rect($left, $bottom + $y * $cellWidth,
						int(@loci) * $cellWidth, $cellWidth);
				}

			}
		}
		$ps->fill();
		$ps->grestore();
	}

	# print values
	$ps->gsave();
	for ($x = 0; $x < @loci; $x++) {
		for ($y = 0; $y < @loci; $y++) {
			next if ($x == $y);
			my $y1 = int(@loci) - $y - 1;
			if ($x < $y1) {
				printDot($ps, $x, $y, fractionFromOR($c->{dots}, $mx->{$loci[$x]}{$loci[$y1]}),
					$c);
			} else {
				printDot($ps, $x, $y, - fractionFromOR($c->{dots},
					$mn->{$loci[$y1]}{$loci[$x]}? (1 / $mn->{$loci[$y1]}{$loci[$x]}): 0), $c);
			}
		}
	}
	$ps->grestore();

	# print legend
	my $fontSize = $c->{font}{size};
	$ps->setfont($c->{font}{name}, $fontSize);
	my $cw = $c->{cellWidth};
	for ($x = 0; $x < @loci; $x++) {
		$ps->showRyC($loci[int(@loci) - $x - 1], $c->{leftBorder},
			$c->{bottomBorder} + $x * $cw + $cw/2);

		$ps->showRotated($loci[$x], $c->{leftBorder} + $x * $cw + $cw/2
			+ 0.3 * $fontSize,
			$c->{bottomBorder} + int(@loci) * $cw + $cw/2,
			$c->{labelRotation});
	}
	# separating line
	$ps->strokeline($left, $bottom + int(@loci) * $cellWidth,
		 $left + int(@loci) * $cellWidth, $bottom);

	drawLegend($ps, $c, int(@loci));
}

sub printPostscript1YI { my ($ps, $c, $mx, $y, $inverse) = @_;
	my @loci = toList(union([keys %{$mx}], [keys %{$mn}]));
	return if (!@loci);
	@loci = @allLoci if ($allLoci);
	@loci = sort { sortTextNumber($a, $b) } @loci;
	print STDERR "Count of Loci ", int(@loci), "\n";

	$ps->setgray(0);

	my ($x);
	# print values
	for ($x = 0; $x < @loci; $x++) {
		next if ( ($inverse && !$mx->{$loci[$x]}) );
		printDot($ps, $x, 0,
			fractionFromOR($c->{dots}, $inverse? (1/$mx->{$loci[$x]}): $mx->{$loci[$x]})
			* ($inverse? -1: 1),
		$c, $y);
	}

	# print legend
	my $fontSize = $c->{font}{size};
	$ps->setfont($c->{font}{name}, $fontSize);
	my $cw = $c->{cellWidth};
	for ($x = 0; $x < @loci; $x++) {
		$ps->showRotated($loci[$x], $c->{leftBorder} + $x * $cw + $cw/2
			+ 0.3 * $fontSize,
			$c->{bottomBorder} + $y + $cw + $cw/2,
			$c->{labelRotation});
	}

}

sub printPostscript1 { my ($ps, $c, $mx, $mn) = @_;
	printPostscript1YI($ps, $c, $mx, $c->{plainOffset}, 0);
	printPostscript1YI($ps, $c, $mn, $c->{inverseOffset}, 1);

	my @loci = toList(union([keys %{$mx}], [keys %{$mn}]));
	@loci = @allLoci if ($allLoci);
	drawLegend($ps, $c, int(@loci));
}

sub prepareDots { my ($c, $gc) = @_;
	$c->{useLogScale} = $gc->{useLogScale};
	$c->{maxOR} = $gc->{maxOR} + 0.0 if (defined($gc->{maxOR}));
	$c->{fct} = \&printDotXY;
	if (defined($gc->{colorsORgreaterOne})) {
		$c->{fct} = \&printColorDotXY;
		mergeDict2dict(\%{$gc->{useColors}.'Colors'}, $c); # <o> obsolete
	}
}

sub printPostscript { my ($c, $mx, $mn, $outputPath) = @_;
	my $depth = dhashdepth($mx);
	print STDERR "Depth: $depth\n";

	my $outputHandle = \*STDOUT;
	if (defined($outputPrefix)) {
		print STDERR "Printing to file: $outputPath\n";
		open(PS_OUTPUTFILE, ">$outputPath");
		$outputHandle = \*PS_OUTPUTFILE;
	}
	my $ps = Postscript->newWithProperties({file => $outputHandle});

	SWITCH: {
		$depth == 1 && do {
			prepareDots($c->{dots} = $c->{orBounds1}, $c);
			printPostscript1($ps, $c, $mx, $mn);
		last SWITCH; };
		$depth == 2 && do {
			prepareDots($c->{dots} = $c->{orBounds2}, $c);
			printPostscript2($ps, $c, $mx, $mn); last SWITCH;
		};
	}
	$ps->close();
	close(PS_OUTPUTFILE) if (defined($outputPath));
}

@defaultPaths = ( ".", "$ENV{HOME}/Dokumente/Genetics/configs",
	"$ENV{HOME}/Documents/Genetics/configs",
	"$ENV{MFCONFIGPATH}");

#main $#ARGV @ARGV %ENV
#	$result = GetOptions('simple' => \$defUndef, 'filter=s' => \$name,
#		'int=i' => \$int, 'onOff!' => \$flag);
	$scheme = 'default';
	$result = GetOptions('help' => \$help, 'h' => \$help, 'postscript' => \$ps,
		logarithmic => \$psCLIConfig{useLogScale}, 'maxOR=i' => \$psCLIConfig{maxOR},
		noGrid => \$psCLIConfig{noGrid}, 'useColors=s' => \$psCLIConfig{useColors},
		'o=s' => \$outputPrefix,
		ree => \$visualizeRee,
		inverse => \$inverseColors,
		allLoci => \$allLoci,
		'scheme=s' => \$scheme,
		drawSingleAlleles => \$psCLIConfig{single},
		'input=s' => \$inputPath
	);

	if ($help || !$result)
	{	printf("USAGE: %s [-o outputPrefix] [--allLoci] [--ree] [--postscript]\n".
			"\tPostscript options: ".
			"--logarithmic --maxOR --noGrid --useColors name\n".
			"\tcolor names: blueYellow redGreen\n",
			($0 =~ m{/?([^/]*)$}o));
		exit(!$result);
	}

	$config = propertyFromString(readFileFirstLocation("mfVisualize.cfg", @defaultPaths));
	if (defined($config->{$scheme})) {
		$psConfig = {%{$config->{default}}};	#<!> global variable
		mergeDict2dict($config->{$scheme}, $psConfig);
		mergeDict2dict(\%psCLIConfig, $psConfig);
	}

	$input = defined($inputPath)? readFile($inputPath): readFileHandle(\*STDIN);
	@allLoci = split(/ /, ($input =~ m{^Loci under scrutiny: (.*)$}om)[0]);
	%models = ($input =~ m{Model: (.*?)\n.*?\n(.*?)(?=Model|$)}osg);
	# this is hardcoded to mffinder output
	$field = optionOn($visualizeRee)? 9: 7;

	foreach $model (keys %models) {
		Log("Scrutinizing model: $model");
		($mx, $mn) = scrutinizeInteraction($psConfig, $models{$model}, $field);
		if ($ps) {
			printPostscript($psConfig, $mx, $mn, "$outputPrefix-$model.ps");
		} else {
			printMatrices($psConfig, $mx, $mn);
		}
		exit(0) if (!defined($outputPrefix));
	}

exit(0);
